@@ -49,9 +49,8 @@ inline double type_epsilon<double>()
 
 TEST(hipfftTest, Create1dPlan)
 {
-    hipfftHandle plan = hipfft_params::INVALID_PLAN_HANDLE;
-    ASSERT_EQ(hipfftCreate(&plan), HIPFFT_SUCCESS);
-    size_t length = 1024;
+    hipfftHandle plan   = hipfft_params::INVALID_PLAN_HANDLE;
+    size_t       length = 1024;
     ASSERT_EQ(hipfftPlan1d(&plan, length, HIPFFT_C2C, 1), HIPFFT_SUCCESS);
 
     ASSERT_EQ(hipfftDestroy(plan), HIPFFT_SUCCESS);
@@ -181,7 +180,6 @@ TEST(hipfftTest, CreatePlanMany64)
     long long int const nY                 = 128;
     long long int const nZ                 = 23;
     long long int       n[3]               = {nX, nY, nZ};
-    long long int       n_invalid[3]       = {nX, -nY, nZ};
     long long int       inembed[3]         = {nX, nY, nZ};
     long long int const istride            = 1;
     long long int const idist              = nX * nY * nZ;
@@ -218,7 +216,10 @@ TEST(hipfftTest, CreatePlanMany64)
         ASSERT_EQ(hipfftDestroy(plan_valid), HIPFFT_SUCCESS);
     }
 
-    // invalid data layout (n array has a negative entry)
+    // invalid data layout (n array has a negative entry). only test rocFFT
+    // backend, since it's more strict
+#ifdef __HIP_PLATFORM_AMD__
+    long long int n_invalid[3] = {nX, -nY, nZ};
     {
         hipfftHandle plan_invalid_1 = hipfft_params::INVALID_PLAN_HANDLE;
         ASSERT_EQ(hipfftCreate(&plan_invalid_1), HIPFFT_SUCCESS);
@@ -238,6 +239,7 @@ TEST(hipfftTest, CreatePlanMany64)
         ASSERT_EQ(hipfftSetAutoAllocation(plan_invalid_1, 0), HIPFFT_SUCCESS);
         ASSERT_EQ(hipfftDestroy(plan_invalid_1), HIPFFT_SUCCESS);
     }
+#endif
 
     // invalid data layout (onembed array has a negative entry)
     {
@@ -560,6 +562,7 @@ TEST(hipfftTest, RunR2C)
     nrmse /= maxv;
 
     EXPECT_TRUE(nrmse < type_epsilon<double>());
+    fftw_destroy_plan(ref_p);
     fftw_free(ref_out);
 }
 
@@ -637,5 +640,6 @@ TEST(hipfftTest, OutplaceOnly)
     nrmse /= maxv;
 
     ASSERT_TRUE(nrmse < type_epsilon<double>());
+    fftw_destroy_plan(ref_p);
     fftw_free(ref_out);
 }
