@@ -28,6 +28,10 @@
 #include <string>
 #include <vector>
 
+#ifdef HIPFFT_MPI_ENABLE
+#include "hipfft/hipfftMp.h"
+#endif
+
 #include "../../../shared/arithmetic.h"
 #include "../../../shared/gpubuf.h"
 #include "../../../shared/ptrdiff.h"
@@ -256,10 +260,19 @@ struct hipfftHandle_t
 
     // Due to hipExec** compatibility to cuFFT, we have to reserve all 4 types
     // rocfft handle separately here.
-    rocfft_plan           ip_forward          = nullptr;
-    rocfft_plan           op_forward          = nullptr;
-    rocfft_plan           ip_inverse          = nullptr;
-    rocfft_plan           op_inverse          = nullptr;
+    rocfft_plan ip_forward = nullptr;
+    rocfft_plan op_forward = nullptr;
+    rocfft_plan ip_inverse = nullptr;
+    rocfft_plan op_inverse = nullptr;
+
+    // return true if the plans have been initialized - hipfftCreate
+    // merely allocates a handle and a hipfftMakePlan* API initializes
+    // them.
+    bool initialized() const
+    {
+        return ip_forward || op_forward || ip_inverse || op_inverse;
+    }
+
     rocfft_execution_info info                = nullptr;
     void*                 workBuffer          = nullptr;
     size_t                workBufferSize      = 0;
@@ -287,6 +300,13 @@ struct hipfftHandle_t
     // brick decomposition for multi-device transforms
     std::vector<hipfft_brick> inBricks;
     std::vector<hipfft_brick> outBricks;
+    // hipFFT will decompose the problem across multiple devices in a
+    // single process (i.e. via hipfftXtSetGPUs)
+    bool singleProcMultiDevice = false;
+
+    // multi-processing communicator
+    rocfft_comm_type comm_type   = rocfft_comm_none;
+    void*            comm_handle = nullptr;
 };
 
 struct hipfft_plan_description_t
@@ -309,6 +329,20 @@ struct hipfft_plan_description_t
     }
 };
 
+static inline hipfftResult handle_exception() noexcept
+try
+{
+    throw;
+}
+catch(hipfftResult e)
+{
+    return e;
+}
+catch(...)
+{
+    return HIPFFT_INTERNAL_ERROR;
+}
+
 hipfftResult hipfftPlan1d(hipfftHandle* plan, int nx, hipfftType type, int batch)
 try
 {
@@ -318,13 +352,9 @@ try
 
     return hipfftMakePlan1d(*plan, nx, type, batch, nullptr);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftPlan2d(hipfftHandle* plan, int nx, int ny, hipfftType type)
@@ -337,13 +367,9 @@ try
 
     return hipfftMakePlan2d(*plan, nx, ny, type, nullptr);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftPlan3d(hipfftHandle* plan, int nx, int ny, int nz, hipfftType type)
@@ -356,13 +382,9 @@ try
 
     return hipfftMakePlan3d(*plan, nx, ny, nz, type, nullptr);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftPlanMany(hipfftHandle* plan,
@@ -385,13 +407,9 @@ try
     return hipfftMakePlanMany(
         *plan, rank, n, inembed, istride, idist, onembed, ostride, odist, type, batch, nullptr);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftPlanMany64(hipfftHandle*  plan,
@@ -414,13 +432,9 @@ try
     return hipfftMakePlanMany64(
         *plan, rank, n, inembed, istride, idist, onembed, ostride, odist, type, batch, nullptr);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftMakePlan_internal(hipfftHandle               plan,
@@ -711,8 +725,11 @@ hipfftResult hipfftMakePlan_internal(hipfftHandle               plan,
         plan->oDist = oDist;
     }
 
-    // problem dimensions and strides are known, set up the bricks for multi-GPU
-    set_io_bricks(plan->inLength, plan->outLength, plan->batch, plan->inBricks, plan->outBricks);
+    // problem dimensions and strides are known, set up the bricks
+    // for single-proc multi-GPU
+    if(plan->singleProcMultiDevice)
+        set_io_bricks(
+            plan->inLength, plan->outLength, plan->batch, plan->inBricks, plan->outBricks);
 
     // create fields for the bricks
     if(!plan->inBricks.empty())
@@ -786,14 +803,21 @@ hipfftResult hipfftMakePlan_internal(hipfftHandle               plan,
 
     if(plan->scale_factor != 1.0)
     {
-        // scale factor requires a rocfft plan description, but
-        // rocfft plan descriptions might not have been created yet
         for(auto rocfft_desc : {ip_forward_desc, op_forward_desc, ip_inverse_desc, op_inverse_desc})
         {
             rocfft_plan_description_set_scale_factor(rocfft_desc, plan->scale_factor);
         }
     }
 
+    // set comm handle on the plans
+    if(plan->comm_type != rocfft_comm_none)
+    {
+        for(auto rocfft_desc : {ip_forward_desc, op_forward_desc, ip_inverse_desc, op_inverse_desc})
+        {
+            rocfft_plan_description_set_comm(rocfft_desc, plan->comm_type, plan->comm_handle);
+        }
+    }
+
     // count the number of plans that got created - it's possible to
     // have parameters that are valid for out-place but not for
     // in-place, so some of these rocfft_plan_creates could
@@ -869,6 +893,11 @@ hipfftResult hipfftMakePlan_internal(hipfftHandle               plan,
         }
     }
 
+    if(workSize != nullptr)
+        *workSize = workBufferSize;
+
+    plan->workBufferSize = workBufferSize;
+
     if(workBufferSize > 0)
     {
         if(plan->autoAllocate)
@@ -886,11 +915,6 @@ hipfftResult hipfftMakePlan_internal(hipfftHandle               plan,
         }
     }
 
-    if(workSize != nullptr)
-        *workSize = workBufferSize;
-
-    plan->workBufferSize = workBufferSize;
-
     rocfft_plan_description_destroy(ip_forward_desc);
     rocfft_plan_description_destroy(op_forward_desc);
     rocfft_plan_description_destroy(ip_inverse_desc);
@@ -918,13 +942,9 @@ try
     *plan = h;
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftExtPlanScaleFactor(hipfftHandle plan, double scalefactor)
@@ -935,13 +955,9 @@ try
     plan->scale_factor = scalefactor;
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult
@@ -964,13 +980,9 @@ try
     return hipfftMakePlan_internal(
         plan, 1, lengths, iotype, number_of_transforms, desc, workSize, false);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftMakePlan2d(hipfftHandle plan, int nx, int ny, hipfftType type, size_t* workSize)
@@ -993,13 +1005,9 @@ try
     return hipfftMakePlan_internal(
         plan, 2, lengths, iotype, number_of_transforms, desc, workSize, false);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult
@@ -1024,13 +1032,9 @@ try
     return hipfftMakePlan_internal(
         plan, 3, lengths, iotype, number_of_transforms, desc, workSize, false);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 template <typename T>
@@ -1160,13 +1164,9 @@ try
     return hipfftMakePlanMany_internal<int>(
         plan, rank, n, inembed, istride, idist, onembed, ostride, odist, iotype, batch, workSize);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftMakePlanMany64(hipfftHandle   plan,
@@ -1189,13 +1189,9 @@ try
     return hipfftMakePlanMany_internal<long long int>(
         plan, rank, n, inembed, istride, idist, onembed, ostride, odist, iotype, batch, workSize);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftEstimate1d(int nx, hipfftType type, int batch, size_t* workSize)
@@ -1205,13 +1201,9 @@ try
     hipfftResult ret  = hipfftGetSize1d(plan, nx, type, batch, workSize);
     return ret;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftEstimate2d(int nx, int ny, hipfftType type, size_t* workSize)
@@ -1221,13 +1213,9 @@ try
     hipfftResult ret  = hipfftGetSize2d(plan, nx, ny, type, workSize);
     return ret;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftEstimate3d(int nx, int ny, int nz, hipfftType type, size_t* workSize)
@@ -1237,13 +1225,9 @@ try
     hipfftResult ret  = hipfftGetSize3d(plan, nx, ny, nz, type, workSize);
     return ret;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftEstimateMany(int        rank,
@@ -1264,13 +1248,9 @@ try
         plan, rank, n, inembed, istride, idist, onembed, ostride, odist, type, batch, workSize);
     return ret;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult
@@ -1289,13 +1269,9 @@ try
 
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftGetSize2d(hipfftHandle plan, int nx, int ny, hipfftType type, size_t* workSize)
@@ -1313,13 +1289,9 @@ try
 
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult
@@ -1338,13 +1310,9 @@ try
 
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftGetSizeMany(hipfftHandle plan,
@@ -1369,13 +1337,9 @@ try
 
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftGetSizeMany64(hipfftHandle   plan,
@@ -1400,13 +1364,9 @@ try
 
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftGetSize(hipfftHandle plan, size_t* workSize)
@@ -1415,13 +1375,9 @@ try
     *workSize = plan->workBufferSize;
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftSetAutoAllocation(hipfftHandle plan, int autoAllocate)
@@ -1431,13 +1387,9 @@ try
         plan->autoAllocate = bool(autoAllocate);
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftSetWorkArea(hipfftHandle plan, void* workArea)
@@ -1456,13 +1408,9 @@ try
     }
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 // Find the specific plan to execute - check placement and direction
@@ -1520,13 +1468,9 @@ try
     }
     return HIPFFT_EXEC_FAILED;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftExecR2C(hipfftHandle plan, hipfftReal* idata, hipfftComplex* odata)
@@ -1534,13 +1478,9 @@ try
 {
     return hipfftExecForward(plan, idata, odata);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftExecC2R(hipfftHandle plan, hipfftComplex* idata, hipfftReal* odata)
@@ -1548,13 +1488,9 @@ try
 {
     return hipfftExecBackward(plan, idata, odata);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftExecZ2Z(hipfftHandle         plan,
@@ -1572,13 +1508,9 @@ try
     }
     return HIPFFT_EXEC_FAILED;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftExecD2Z(hipfftHandle plan, hipfftDoubleReal* idata, hipfftDoubleComplex* odata)
@@ -1586,13 +1518,9 @@ try
 {
     return hipfftExecForward(plan, idata, odata);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftExecZ2D(hipfftHandle plan, hipfftDoubleComplex* idata, hipfftDoubleReal* odata)
@@ -1600,13 +1528,9 @@ try
 {
     return hipfftExecBackward(plan, idata, odata);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftSetStream(hipfftHandle plan, hipStream_t stream)
@@ -1615,13 +1539,9 @@ try
     ROC_FFT_CHECK_INVALID_VALUE(rocfft_execution_info_set_stream(plan->info, stream));
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftDestroy(hipfftHandle plan)
@@ -1651,13 +1571,9 @@ try
 
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftGetVersion(int* version)
@@ -1688,13 +1604,9 @@ try
     *version = std::stoi(result.str());
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftGetProperty(hipfftLibraryPropertyType type, int* value)
@@ -1718,13 +1630,9 @@ try
 
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtSetCallback(hipfftHandle         plan,
@@ -1818,13 +1726,9 @@ try
         return HIPFFT_INVALID_VALUE;
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtClearCallback(hipfftHandle plan, hipfftXtCallbackType cbtype)
@@ -1832,13 +1736,9 @@ try
 {
     return hipfftXtSetCallback(plan, nullptr, cbtype, nullptr);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult
@@ -1881,13 +1781,9 @@ try
         return HIPFFT_INVALID_VALUE;
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtMakePlanMany(hipfftHandle   plan,
@@ -1911,13 +1807,9 @@ try
     return hipfftMakePlanMany_internal<long long int>(
         plan, rank, n, inembed, istride, idist, onembed, ostride, odist, iotype, batch, workSize);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtGetSizeMany(hipfftHandle   plan,
@@ -1947,13 +1839,9 @@ try
     HIP_FFT_CHECK_AND_RETURN(hipfftDestroy(p));
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtExec(hipfftHandle plan, void* input, void* output, int direction)
@@ -1974,13 +1862,9 @@ try
 
     return hipfftExec(plan_ptr, plan->info, input, output);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtSetGPUs(hipfftHandle plan, int count, int* gpus)
@@ -2003,15 +1887,13 @@ try
         plan->outBricks[i].device = gpus[i];
     }
 
+    plan->singleProcMultiDevice = true;
+
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 // get number of bytes used for elements of a given hipDataType
@@ -2101,13 +1983,9 @@ try
     *desc                = lib_desc.release();
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_ALLOC_FAILED;
+    return handle_exception();
 }
 
 // collapse contiguous dimensions in the specified length + stride -
@@ -2293,13 +2171,9 @@ try
         throw HIPFFT_INVALID_VALUE;
     }
 }
-catch(hipfftResult err)
-{
-    return err;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtFree(hipLibXtDesc* desc)
@@ -2317,13 +2191,9 @@ try
     delete desc;
     return HIPFFT_SUCCESS;
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 static hipfftResult hipfftXtExecDescriptorBase(const rocfft_plan&           rplan,
@@ -2355,13 +2225,9 @@ try
 
     return hipfftXtExecDescriptorBase(rplan, plan->info, input, output);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtExecDescriptorR2C(hipfftHandle plan, hipLibXtDesc* input, hipLibXtDesc* output)
@@ -2375,13 +2241,9 @@ try
 
     return hipfftXtExecDescriptorBase(rplan, plan->info, input, output);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtExecDescriptorC2R(hipfftHandle plan, hipLibXtDesc* input, hipLibXtDesc* output)
@@ -2395,13 +2257,9 @@ try
 
     return hipfftXtExecDescriptorBase(rplan, plan->info, input, output);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtExecDescriptorZ2Z(hipfftHandle  plan,
@@ -2418,13 +2276,9 @@ try
 
     return hipfftXtExecDescriptorBase(rplan, plan->info, input, output);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtExecDescriptorD2Z(hipfftHandle plan, hipLibXtDesc* input, hipLibXtDesc* output)
@@ -2438,13 +2292,9 @@ try
 
     return hipfftXtExecDescriptorBase(rplan, plan->info, input, output);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtExecDescriptorZ2D(hipfftHandle plan, hipLibXtDesc* input, hipLibXtDesc* output)
@@ -2458,13 +2308,9 @@ try
 
     return hipfftXtExecDescriptorBase(rplan, plan->info, input, output);
 }
-catch(hipfftResult e)
-{
-    return e;
-}
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
 
 hipfftResult hipfftXtExecDescriptor(hipfftHandle  plan,
@@ -2481,11 +2327,110 @@ try
 
     return hipfftXtExecDescriptorBase(rplan, plan->info, input, output);
 }
-catch(hipfftResult e)
+catch(...)
 {
-    return e;
+    return handle_exception();
+}
+
+#ifdef HIPFFT_MPI_ENABLE
+static rocfft_comm_type hipfftMpCommTypeToRocfftCommType(hipfftMpCommType_t hipfft_type)
+{
+    switch(hipfft_type)
+    {
+    case HIPFFT_COMM_MPI:
+        return rocfft_comm_mpi;
+    case HIPFFT_COMM_NONE:
+        return rocfft_comm_none;
+    }
+    throw HIPFFT_INVALID_VALUE;
+}
+
+hipfftResult hipfftMpAttachComm(hipfftHandle plan, hipfftMpCommType comm_type, void* comm_handle)
+try
+{
+    // comm must be known before plans are actually constructed
+    if(!plan || plan->initialized())
+        return HIPFFT_INVALID_PLAN;
+
+    plan->comm_type   = hipfftMpCommTypeToRocfftCommType(comm_type);
+    plan->comm_handle = comm_handle;
+    return HIPFFT_SUCCESS;
 }
 catch(...)
 {
-    return HIPFFT_INTERNAL_ERROR;
+    return handle_exception();
 }
+
+hipfftResult hipfftXtSetDistribution(hipfftHandle         plan,
+                                     int                  rank,
+                                     const long long int* input_lower,
+                                     const long long int* input_upper,
+                                     const long long int* output_lower,
+                                     const long long int* output_upper,
+                                     const long long int* input_stride,
+                                     const long long int* output_stride)
+try
+{
+    // distribution must be set before plans are actually constructed
+    if(!plan || plan->initialized())
+        return HIPFFT_INVALID_PLAN;
+
+    // one brick on this rank for each of input and output
+    plan->inBricks.resize(1);
+    plan->outBricks.resize(1);
+
+    auto setBrick = [=](hipfft_brick&        b,
+                        const long long int* lower,
+                        const long long int* upper,
+                        const long long int* stride) {
+        // init brick for FFT dimensions + batch dimension
+        b.field_lower.resize(rank + 1);
+        b.field_upper.resize(rank + 1);
+        b.brick_stride.resize(rank + 1);
+
+        // copy row-major coordinates and strides to column-major brick info
+        std::reverse_iterator<const long long int*> lower_rbegin(lower + rank);
+        std::reverse_iterator<const long long int*> lower_rend(lower);
+        std::copy(lower_rbegin, lower_rend, b.field_lower.begin());
+        std::reverse_iterator<const long long int*> upper_rbegin(upper + rank);
+        std::reverse_iterator<const long long int*> upper_rend(upper);
+        std::copy(upper_rbegin, upper_rend, b.field_upper.begin());
+        std::reverse_iterator<const long long int*> stride_rbegin(stride + rank);
+        std::reverse_iterator<const long long int*> stride_rend(stride);
+        std::copy(stride_rbegin, stride_rend, b.brick_stride.begin());
+
+        // hipFFT only supports batch-1 distributed FFTs, so set lower
+        // + upper + stride for batch dimension
+        b.field_lower.back()  = 0;
+        b.field_upper.back()  = 1;
+        b.brick_stride.back() = 0;
+
+        (void)hipGetDevice(&b.device);
+    };
+
+    setBrick(plan->inBricks.front(), input_lower, input_upper, input_stride);
+    setBrick(plan->outBricks.front(), output_lower, output_upper, output_stride);
+    return HIPFFT_SUCCESS;
+}
+catch(...)
+{
+    return handle_exception();
+}
+
+hipfftResult hipfftXtSetSubformatDefault(hipfftHandle      plan,
+                                         hipfftXtSubFormat subformat_forward,
+                                         hipfftXtSubFormat subformat_inverse)
+try
+{
+    // formats must be set before plans are actually constructed
+    if(!plan || plan->initialized())
+        return HIPFFT_INVALID_PLAN;
+
+    return HIPFFT_NOT_IMPLEMENTED;
+}
+catch(...)
+{
+    return handle_exception();
+}
+
+#endif
